/***************************************************************************
 *   Copyright (C) 2006 by Gerd v. Egidy                                   *
 *   gve@intra2net.com                                                     *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string>
#include <sstream>

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/serialization/serialization.hpp>

#include <boost/bind.hpp>

#include "command_client.hxx"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

using namespace std;

namespace libt2n
{

command_client::command_client(client_connection& _c, long long _command_timeout_usec, long long _hello_timeout_usec)
    : c(_c)
{
    command_timeout_usec=_command_timeout_usec;
    hello_timeout_usec=_hello_timeout_usec;

    // for reconnects
    c.add_callback(new_connection,bind(&command_client::read_hello, boost::ref(*this)));

    read_hello();
}

std::string command_client::read_packet(const long long &usec_timeout)
{
    string resultpacket;
    bool got_packet=false;
    long long my_timeout=usec_timeout;
    while(!(got_packet=c.get_packet(resultpacket)) && my_timeout > 0  && !c.is_closed())
        c.fill_buffer(my_timeout,&my_timeout);

    if (!got_packet)
        throw t2n_transfer_error("timeout exceeded");

    return resultpacket;
}

void command_client::read_hello()
{
    istringstream hello(read_packet(hello_timeout_usec));

    char chk[5];
    hello.read(chk,4);
    chk[4]=0;
    if (hello.fail() || hello.eof() || string("T2Nv") != chk)
        throw t2n_version_mismatch("illegal hello received (T2N)");

    int prot_version;
    hello >> prot_version;
    if (hello.fail() || hello.eof() || prot_version != PROTOCOL_VERSION)
        throw t2n_version_mismatch("not compatible with the server protocol version");

    hello.read(chk,1);
    if (hello.fail() || hello.eof() || chk[0] != ';')
        throw t2n_version_mismatch("illegal hello received (1. ;)");

    hello.read(chk,4);
    if (hello.fail() || hello.eof() || *((int*)chk) != 1)
        throw t2n_version_mismatch("host byte order not matching");

    hello.read(chk,1);
    if (hello.fail() || hello.eof() || chk[0] != ';')
        throw t2n_version_mismatch("illegal hello received (2. ;)");
}

void command_client::send_command(command* cmd, result_container &res)
{
    ostringstream ofs;
    command_container cc(cmd);
    boost::archive::binary_oarchive oa(ofs);

    try
    {
        oa << cc;
    }
    catch(boost::archive::archive_exception &e)
    {
        ostringstream msg;
        msg << "archive_exception while serializing on client-side, code " << e.code << " (" << e.what() << ")";
        throw t2n_serialization_error(msg.str());
    }
    catch(...)
        { throw; }

    std::ostream* ostr;
    if ((ostr=c.get_logstream(fulldebug))!=NULL)
    {
        (*ostr) << "sending command, decoded data: " << std::endl;
        boost::archive::xml_oarchive xo(*ostr);
        xo << BOOST_SERIALIZATION_NVP(cc);
    }

    c.write(ofs.str());

    istringstream ifs(read_packet(command_timeout_usec));
    boost::archive::binary_iarchive ia(ifs);

    try
    {
        ia >> res;
    }
    catch(boost::archive::archive_exception &e)
    {
        ostringstream msg;
        msg << "archive_exception while deserializing on client-side, code " << e.code << " (" << e.what() << ")";
        throw t2n_serialization_error(msg.str());
    }
    catch(...)
        { throw; }

    if ((ostr=c.get_logstream(fulldebug))!=NULL)
    {
        (*ostr) << "received result, decoded data: " << std::endl;
        boost::archive::xml_oarchive xo(*ostr);
        xo << BOOST_SERIALIZATION_NVP(res);
    }
}

}
